import io
import os
import math
import struct
from collections import defaultdict
from tempfile import TemporaryDirectory
import sys

import msgpack
import numpy as np

import homcloud.alpha_shape3
import homcloud.alpha_shape2
from homcloud.index_map import IndexMapForAlphaFiltration
from homcloud.filtration import Filtration
import homcloud.utils_base as utils

def dict_simplices(simplices):
    return {simplex.key: simplex for simplex in simplices}

MINUS_INF = -sys.float_info.max


class AlphaShape(object):
    def create_filtration(self, no_square):
        simplices = [Simplex(alpha_simplex, i, no_square)
                     for (i, alpha_simplex) in enumerate(self.simplices())]
        return AlphaFiltration(self.coordinates_of_vertices(), simplices,
                               dict_simplices(simplices), self.dim())

    def coordinates_of_vertices(self):
        return self.coordinates

class AlphaShape3(homcloud.alpha_shape3.AlphaShape3, AlphaShape):
    def __init__(self, points, weighted, rel_homology):
        super().__init__(points, weighted, rel_homology)
        self.coordinates = points[:, 0:3]

    def simplices(self):
        return self.vertices() + self.edges() + self.facets() + self.cells()

    @staticmethod
    def dim():
        return 3

class AlphaShape2(homcloud.alpha_shape2.AlphaShape2, AlphaShape):
    def __init__(self, points, weighted, rel_homology):
        super().__init__(points, weighted, rel_homology)
        self.coordinates = points[:, 0:2]

    def simplices(self):
        return self.vertices() + self.edges() + self.faces()

    @staticmethod
    def dim():
        return 2

class Simplex(object):
    """A class representing simplex
    This class wraps PC2diphacomplex.alpha_simplex.{Vertex, Cell, Edge, Facet}.
    """
    def __init__(self, alpha_simplex, index, no_square=False):
        self.alpha_simplex = alpha_simplex
        self.index = index # NOTE: This index is different from vertex's index
        self.key = frozenset(v.index for v in self.alpha_simplex.vertices())
        self.no_square = no_square
        self.true_birth_radius = self.birth_radius = \
                                 self.normalize_radius(alpha_simplex.birth_radius)

    def boundary_keys(self):
        """Return list of frozensets of vertices of indices of boundary simplices"""
        if self.isvertex():
            return []
        return [self.key.difference([el]) for el in self.key]

    # def true_birth_radius(self):
    #     """Return "true" birth radius.

    #     birth_radius returns the renamed value when indexing is used.
    #     """
    #     return self.normalize_radius(self.alpha_simplex.birth_radius)

    def isvertex(self):
        """Return True if the simplex wraps a vertex.
        """
        return (isinstance(self.alpha_simplex, homcloud.alpha_shape2.Vertex)
                or isinstance(self.alpha_simplex, homcloud.alpha_shape3.Vertex))

    def normalize_radius(self, r):
        if self.no_square:
            return math.copysign(math.sqrt(abs(r)), r)
        else:
            return r

    def group_name(self):
        """Return group name as an integer

        Return None if the simplex does not belong to any group. 
        A simplex belongs to a group if it satisfies the following two conditions:

        * All group_name of its vertices are the same.
        * Any group_name of its vertices is not -1.
        """
        group_names_of_vertice = [v.group_name for v in self.alpha_simplex.vertices()]
        first_name = group_names_of_vertice[0]
        if first_name == -1:
            return None
        if all(name == first_name for name in group_names_of_vertice):
            return first_name
        else:
            return None

    @property
    def dim(self):
        """Return the dimension of the simplex"""
        return len(self.key) - 1

class AlphaFiltration(Filtration):
    def __init__(self, points, simplices, dict_simplices, dim):
        self.points = points
        self.simplices = simplices
        self.dict_simplices = dict_simplices
        self.dim = dim
        self.index_map = None

    def indexize(self):
        """Index-ize self
        self.simplices is modified (birth_radius is replaced
        by index value) in this method.

        An object of IndexMapForAlphaFiltration is set to self.index_map.
        """
        ordered_simplices = self.rename_birth_radii(self.simplices)
        self.index_map = self.create_index_map(ordered_simplices, self.points, self.dim)


    def write_dipha_complex(self, output):
        """"Write dipha bondary matrix data to file"""
        def write_dipha_header(num_simplices):
            """Write magic number, type, boundary/coboundary, num_simplices, and dimension"""
            output.write(struct.pack("qqqqq", 8067171840, 0, 0, num_simplices, self.dim))

        def write_dimensions():
            """Write dimensions of each simplex"""
            for simplex in self.simplices:
                output.write(struct.pack("q", simplex.dim))

        def write_birth_radii():
            """Write birth radii for each simplex"""
            for simplex in self.simplices:
                output.write(struct.pack("d", simplex.birth_radius))

        def write_boundary_map_sizes():
            n = 0
            for simplex in self.simplices:
                output.write(struct.pack("q", n))
                n += len(simplex.boundary_keys())
            output.write(struct.pack("q", n)) # Write the number of all boundary elements

        def write_boundary_map():
            for simplex in self.simplices:
                for key in simplex.boundary_keys():
                    output.write(struct.pack("q", self.dict_simplices[key].index))

        write_dipha_header(len(self.simplices))
        write_dimensions()
        write_birth_radii()
        write_boundary_map_sizes()
        write_boundary_map()

    def write_indexed_complex(self, output):
        """Write the dipha bondary matrix data with index-map information to output.
        """
        complex_buf = io.BytesIO()
        self.write_dipha_complex(complex_buf)

        output.write(msgpack.packb({
            "dipha-data": complex_buf.getvalue(),
            "index-map": self.index_map.to_dict(),
        }, use_bin_type=True))

    def subsets_simplices(self):
        subsets = defaultdict(list)
        for simplex in self.simplices:
            group_name = simplex.group_name()
            if group_name is not None:
                subsets[group_name].append(simplex)
        return subsets

    def subsets(self):
        def subset(simplices):
            new_simplices = [Simplex(simplex.alpha_simplex, i, simplex.no_square)
                             for (i, simplex) in enumerate(simplices)]
            return AlphaFiltration(self.points, new_simplices,
                                   dict_simplices(new_simplices), self.dim)

        return {
            group_name: subset(simplices)
            for (group_name, simplices) in self.subsets_simplices().items()
        }

    def isacyclic(self):
        def read_num_essential_pairs(f):
            num_ess_pairs = 0
            num_pairs = utils.read_diagram_header(f)
            for _ in range(num_pairs):
                pair_d, _, _ = struct.unpack("qdd", f.read(24))
                if pair_d < 0:
                    num_ess_pairs += 1
            return num_ess_pairs

        with TemporaryDirectory() as tmpdir:
            outpath = os.path.join(tmpdir, "tmp.diagram")
            self.compute_diagram_and_save(outpath)
            with open(outpath, "rb") as f:
                return read_num_essential_pairs(f) == 1

    def build_grouped_filtration(self):
        for subset in self.subsets().values():
            for simplex_key in subset.dict_simplices:
                simplex = self.dict_simplices[simplex_key]
                simplex.true_birth_radius = simplex.birth_radius = MINUS_INF

    def check_for_relative_filtration(self):
        for subset in self.subsets().values():
            if not subset.isacyclic():
                raise RuntimeError("A subset is not acyclic")

    @staticmethod
    def rename_birth_radii(simplices):
        """Indexing simplices by their birth radius.

        For each simplex in simplices, its birth_radius is set to n where
        its birth_radius is n-th smallest in the set of all birth_radii of simplices.
        This method returns simplices ordered by their birth radii.
        """
        from operator import attrgetter
        table = sorted(simplices, key=attrgetter("birth_radius"))
        for (index, simplex) in enumerate(table):
            simplex.birth_radius = index
        return table

    @staticmethod
    def create_index_map(simplices, points, dim):
        """Return dictionary of indices.
        """
        return IndexMapForAlphaFiltration(
            np.array([simplex.true_birth_radius for simplex in simplices], dtype=float),
            points,
            [list(simplex.key) for simplex in simplices],
            dim
        )

def create_alpha_shape(points, dim, weighted=False, use_relative_homology=False):
    if dim == 2:
        return AlphaShape2(points, weighted, use_relative_homology)
    if dim == 3:
        return AlphaShape3(points, weighted, use_relative_homology)
    raise ValueError("dimension of a point cloud must be 2 or 3")

