import struct

import numpy as np
import msgpack

from homcloud.index_map import IndexMapForBitmap
from homcloud.pict.utils import nd_indices
from homcloud.filtration import Filtration


class BitmapComplex(Filtration):
    """A class representing N-d bitmap complex
    """
    def __init__(self, array, flip_levels_sign=None):
        self.array = array
        self.index_map = None
        self.dim = array.ndim
        self.flip_levels_sign = flip_levels_sign

    def write_dipha_complex(self, f):
        """Write a picture to f with the dipha format.
        """
        f.write(struct.pack("qq", 8067171840, 1))
        f.write(struct.pack("qq", self.array.size, self.array.ndim))
        for g in reversed(self.array.shape):
            f.write(struct.pack("q", g))

        for pixel in self.array.flatten():
            f.write(struct.pack("d", pixel))

    def dipha_complex_size(self):
        header_size = 8*4
        ndim_size = 8*self.array.ndim
        data_size = 8*self.array.size
        return header_size + ndim_size + data_size

    def set_mask(self, mask):
        self.array[mask] = 100000

    def save(self, complex_path, combine_index_map):
        """Save a BitmapComplex object to complex_path.

        Note that self is modified when combine_index_map is True.

        Args:
        complex_path -- Output file path
        combine_index_map -- If True, combine complex and index map into one file.
        """
        with open(complex_path, "wb") as f:
            if combine_index_map:
                self.indexize()
                self.write_indexed_complex(f)
            else:
                self.write_dipha_complex(f)

    def write_indexed_complex(self, output):
        """Write a picture with index-map to output

        This method is available for "indexize"d bitmap.
        """
        if self.index_map is None:
            raise ValueError("The BitmapComplex object is not indexized")
        packer = msgpack.Packer(use_bin_type=True)
        output.write(packer.pack_map_header(2))
        output.write(packer.pack("dipha-data"))
        output.write(struct.pack(">BL", 0xc6, self.dipha_complex_size()))
        self.write_dipha_complex(output)
        output.write(packer.pack("index-map"))
        output.write(packer.pack(self.index_map.to_dict()))

    def indexize(self):
        """Index-ize self.

        An object of IndexMapForBitmap is set to self.index_map.
        """
        indices = nd_indices(self.array.shape)
        key = np.argsort(self.array, axis=None, kind="mergesort")
        bitmap_index = invert_permutation(key).reshape(self.array.shape)
        self.index_map = IndexMapForBitmap(indices[key, :],
                                           self.sign() * self.array.flatten()[key],
                                           self.array.ndim, self.flip_levels_sign)
        self.array = bitmap_index

    def sign(self):
        return -1 if self.flip_levels_sign else 1


def invert_permutation(a):
    """Create the inverse of the permutation given by a.

    If a is not a permutation, the result is indefinite.
    """
    s = np.zeros(a.size, dtype=np.int32)
    i = np.arange(a.size, dtype=np.int32)
    np.put(s, a, i)
    return s
