"""A module providing binarization and enlargement of pictures
"""
import math
import argparse
import sys

import numpy as np
from sklearn.neighbors import BallTree, DistanceMetric
from scipy.ndimage.morphology import binary_erosion
from scipy.ndimage import distance_transform_cdt, distance_transform_edt

import homcloud.pict.utils as utils
from homcloud.bitmap_complex import BitmapComplex
from homcloud.base import __version__
from homcloud.license import add_argument_for_license


def main(args=None):
    args = args or argument_parser().parse_args()
    pict = utils.load_picture(args.input, filetype=args.type)
    binary_pict = binarize_picture(pict, args.threshold, args.mode, (args.gt, args.lt))
    metric = construct_metric(args.metric, args.weight, args.angle)

    bitmap = distance_bitmap(binary_pict, metric, args.ensmall)

    if args.mask:
        bitmap.set_mask(utils.load_picture(args.mask) > 0)

    if args.combine_index_map:
        bitmap.indexize()

    if args.convert_to_diagram:
        bitmap.compute_diagram_and_save(args.output)
    else:
        bitmap.write_file(args.output)

    return 0


def argument_parser():
    p = argparse.ArgumentParser(description="Construct cubical filtrations from binarized digital pictures using a distance on the pictures")
    p.add_argument("-V", "--version", action="version", version=__version__)
    add_argument_for_binarize_threshold(p)
    p.add_argument("-T", "--type", default="picture",
                   help="input file format (picture(default), text)")
    p.add_argument("--metric", default="manhattan",
                   help="metric used to enlarge binarized image"
                   + "(manhattan, euclidean, etc., default is manhattan) ")
    p.add_argument("-I", "--combine-index-map", default=False, action="store_true",
                   help="combine the index map with the output filtration")
    p.add_argument("-s", "--ensmall", action="store_true", default=False,
                   help="ensmall binarized picture")
    p.add_argument("-w", "--weight", type=float, default=1,
                   help="weight for mahalanobis metric")
    p.add_argument("-a", "--angle", type=float, default=0.0,
                   help="angle for mahalanobis metric")
    p.add_argument("-M", "--mask", default=None,
                   help="mask bitmap (white region is used for mask)")
    p.add_argument("-D", "--convert-to-diagram", default=False, action="store_true",
                   help="call dipha and directly convert to a diagram")
    add_argument_for_license(p)
    p.add_argument("input", help="input file name")
    p.add_argument("output", help="output file name")
    return p


def add_argument_for_binarize_threshold(p):
    p.add_argument("-m", "--mode", default="black-base",
                   help="the way to construct dipha complex (black-base|white-base, "
                   + "default is black-base)")
    p.add_argument("-t", "--threshold", type=float, default=128,
                   help="threshold for binarization (default: 128)")
    p.add_argument("--gt", type=float, default=None, help="lower threshold")
    p.add_argument("--lt", type=float, default=None, help="upper threshold")


def binarize_picture(pict, threshold, mode, bounds):
    """Binarize pict using threshold.

    Args:
    pict -- a gray-scale picture (ndarray)
    threshold -- threshold for binarization
    mode -- "black-base" or "white-base"
    bounds -- a pair of lower and upper bounds
        if both of them are None, the pair of threshold and mode is used
    """
    lower_bound, upper_bound = bounds

    if lower_bound is None and upper_bound is None:
        if mode == "black-base":
            upper_bound = threshold
        elif mode == "white-base":
            lower_bound = threshold
        else:
            raise RuntimeError("Unknown mode {0}".format(mode))

    lower_bound = -float("inf") if lower_bound is None else lower_bound
    upper_bound = float("inf") if upper_bound is None else upper_bound
    return (lower_bound <= pict) & (pict <= upper_bound)


def construct_metric(name, weight, angle):
    if name == "mahalanobis":
        return mahalanobis_metric(weight, angle*math.pi)
    else:
        return name


def mahalanobis_metric(weight, angle):
    def rotation_matrix(t):
        from math import cos, sin
        return np.array([[cos(t), sin(t)], [-sin(t), cos(t)]])
    matrix = np.dot(np.dot(rotation_matrix(angle), np.diag([weight, 1/weight])),
                    rotation_matrix(-angle))
    return DistanceMetric.get_metric("mahalanobis", VI=matrix)


def pixel_positions(pict):
    """Return indices of True in 2D boolean ndarray pict
    """
    indices = utils.nd_indices(pict.shape)
    return indices[pict.flatten()]


def boundary_pixels(pict):
    return pixel_positions(
        np.logical_xor(pict, binary_erosion(pict, iterations=1, border_value=1))
    )


DISTANCE_TRANSFORM_FUNCTIONS = {
    "manhattan": lambda pict: distance_transform_cdt(pict, "taxicab"),
    "chebyshev": lambda pict: distance_transform_cdt(pict, "chessboard"),
    "euclidean": distance_transform_edt
}


def distance_bitmap(pict, metric, use_negative_distances):
    """Create a "enlarged/shrinked" BitmapComplex.

    Args:
    pict -- binary picture, it means N-dimensional ndarray whose dtype is bool.
    metric -- Metric used for enlarging and ensmalling
    use_negative_distances -- True for ensmalling
    """
    if metric in DISTANCE_TRANSFORM_FUNCTIONS:
        distance_transform = DISTANCE_TRANSFORM_FUNCTIONS[metric]
        array = distance_transform(~pict)
        if use_negative_distances:
            array[pict] = -distance_transform(pict)[pict]

        return BitmapComplex(array, False)

    return distance_bitmap_generic(pict, metric, use_negative_distances)


def distance_bitmap_generic(pict, metric, use_negative_distances):
    """Generic distance
    """
    array = np.zeros(pict.shape)
    boundary = boundary_pixels(pict)
    balltree = BallTree(boundary, leaf_size=30, metric=metric)
    inverted_pict = ~pict

    def compute_distances(targets):
        (d, _) = balltree.query(targets, k=1)
        return d[:, 0]

    empty_pixels = pixel_positions(inverted_pict)
    distances = compute_distances(empty_pixels)
    np.place(array, inverted_pict, distances)

    if use_negative_distances:
        nonempty_pixels = pixel_positions(pict)
        negative_distances = -compute_distances(nonempty_pixels) - 1
        np.place(array, pict, negative_distances)

    return BitmapComplex(array, False)


if __name__ == "__main__":
    sys.exit(main())
