import argparse
import itertools
import operator
import json

import numpy as np
import msgpack

from homcloud.base import __version__
import homcloud.pict.utils as utils
import homcloud.pict.binarize as binarize
from homcloud.bitmap_complex import BitmapComplex
from homcloud.license import add_argument_for_license

def argument_parser():
    p = argparse.ArgumentParser(description="Compute 0th PH tree and (n-1)-th PH tree")
    p.add_argument("-V", "--version", action="version", version=__version__)
    add_argument_for_license(p)
    p.add_argument("-m", "--mode", required=True,
                   help=("mode (white-base or black-base for binarize," +
                         " superlevel or sublevel for levelset)"))

    for_binarize = p.add_argument_group('for binarize')
    for_binarize.add_argument("-t", "--threshold", type=float, default=128)
    for_binarize.add_argument("--gt", type=float, default=None, help="lower threshold")
    for_binarize.add_argument("--lt", type=float, default=None, help="upper threshold")
    for_binarize.add_argument("-s", "--ensmall", action="store_true", default=False,
                              help="ensmall binarized picture")
    for_binarize.add_argument(
        "--metric", default="manhattan",
        help="metric used to enlarge binarized image (manhattan(default), euclidean, etc.)"
    )
    #for_binarize.add_argument("--matrix", help="not implemnted yet")

    # arguments for input and output
    for_input_output = p.add_argument_group("for input and output")
    for_input_output.add_argument(
        "-T", "--type", default="text_nd",
        help="input data format (text2d, text_nd(default), picture2d, picture3d, npy)"
    )
    for_input_output.add_argument("-O", "--output-type", default="msgpack",
                                  help="output file format (json, msgpack(default))")
    for_input_output.add_argument("-o", "--output", required=True, help="output file")
    p.add_argument("input", nargs="*", help="input files")
    return p

def main(args=None):
    args = args or argument_parser().parse_args()
    pict = utils.load_nd_picture(args.input, args.type)

    if args.mode in ["black-base", "white-base"]:
        binary_pict = binarize.binarize_picture(pict, args.threshold, args.mode,
                                                (args.gt, args.lt))
        bitmap = binarize.distance_bitmap(binary_pict, args.metric, args.ensmall)
    elif args.mode in ["sublevel", "superlevel"]:
        if args.mode == "superlevel":
            pict = -pict.astype(float)
        bitmap = BitmapComplex(pict)
    else:
        raise RuntimeError("invalid mode")

    low_mergetree = LowerMergeTree(bitmap.array)
    low_mergetree.compute()
    up_mergetree = UpperMergeTree(bitmap.array)
    up_mergetree.compute()
    dic = {
        "dim": bitmap.array.ndim,
        "lower": low_mergetree.to_dict(),
        "upper": up_mergetree.to_dict(),
    }
    if args.output_type == "json":
        with open(args.output, "w") as f:
            json.dump(dic, f)
    elif args.output_type == "msgpack":
        with open(args.output, "wb") as f:
            msgpack.dump(dic, f, use_bin_type=True)


class MergeTree(object):
    def __init__(self, pixel_to_node):
        self.pixel_to_node = pixel_to_node
        self.ordered_nodes = sorted(pixel_to_node.ravel(),
                                    key=operator.attrgetter("value"),
                                    reverse=self.reverse_order())
        for (index, node) in enumerate(self.ordered_nodes):
            node.index = index

    def compute(self):
        for node in self.ordered_nodes:
            for adj_node in self.adj_nodes(node):
                if node.index <= adj_node.index:
                    continue
                root1 = node.root()
                root2 = adj_node.root()
                if root1 is root2:
                    continue
                root1, root2 = self.order_nodes(root1, root2)
                root2.set_parent(root1, node.value, node)

    def adj_nodes(self, node):
        base = np.array(node.pixel, dtype=int)
        return [self.pixel_to_node[tuple(v + base)] for v in self.adj_vectors
                if not out_of_area(v + base, self.pixel_to_node.shape)]

    def to_dict(self):
        return {
            "degree": self.degree(),
            "nodes": {str(node.index): node.to_dict(self.is_dual()) for node
                      in self.ordered_nodes if node.is_nontrivial()}
        }

    @staticmethod
    def order_nodes(node1, node2):
        if node1.index < node2.index:
            return node1, node2
        else:
            return node2, node1

    class Node(object):
        def __init__(self, pixel, value, on_boundary):
            self.pixel = pixel
            self.value = value
            self.on_boundary = on_boundary
            self.index = None
            self.parent = self.root_shortcut = None
            self.death_time = None
            self.death_node = None
            self.children = []

        def __repr__(self):
            return "homcloud.pict.tree.MergeTree.Node({}, {})".format(
                self.pixel, self.value
            )

        def root(self):
            if self.parent is None:
                return self
            self.root_shortcut = self.root_shortcut.root()
            return self.root_shortcut

        def set_parent(self, parent, level, death_node):
            self.parent = self.root_shortcut = parent
            self.death_time = level
            self.death_node = death_node
            parent.children.append(self)

        def volume_nodes(self):
            nodes = []
            def visit(node):
                nodes.append(node)
                for child in node.children:
                    visit(child)
            visit(self)
            return nodes

        def is_nontrivial(self):
            return (self.death_time != self.value) and (not self.on_boundary)

        def to_dict(self, dual):
            if dual:
                birth_time = self.death_time
                birth_pixel = self.death_node.pixel
                death_time = self.value
                death_pixel = self.pixel
                parent = None if np.isinf(self.parent.value) else str(self.parent.index)
            else:
                birth_time = self.value
                birth_pixel = self.pixel
                death_time = self.death_time
                death_pixel = self.death_node.pixel if self.death_node else None
                parent = str(self.parent.index) if self.parent else None

            return {
                "id": str(self.index), "parent": parent,
                "birth-time": birth_time, "death-time": death_time,
                "birth-pixel": birth_pixel, "death-pixel": death_pixel,
                "volume": [n.pixel for n in self.volume_nodes()]
            }

    class InfNode(object):
        def __init__(self):
            self.value = np.inf
            self.index = -1
            self.children = []

        def root(self):
            return self


class UpperMergeTree(MergeTree):
    def __init__(self, filtration):
        pixel_to_node = np.empty_like(filtration, dtype=object)
        infnode = MergeTree.InfNode()
        for mindex in np.ndindex(filtration.shape):
            node = MergeTree.Node(mindex, int(filtration[mindex]),
                                  on_boundary(mindex, filtration.shape))
            if node.on_boundary:
                node.set_parent(infnode, None, -1)

            pixel_to_node[mindex] = node
        super().__init__(pixel_to_node)
        self.adj_vectors = dual_adj_vectors(filtration.ndim)

    @staticmethod
    def reverse_order():
        return True

    @staticmethod
    def is_dual():
        return True

    def degree(self):
        return self.pixel_to_node.ndim - 1

class LowerMergeTree(MergeTree):
    def __init__(self, filtration):
        pixel_to_node = np.empty_like(filtration, dtype=object)
        for mindex in np.ndindex(filtration.shape):
            pixel_to_node[mindex] = MergeTree.Node(mindex, int(filtration[mindex]), False)

        super().__init__(pixel_to_node)
        self.adj_vectors = adj_vectors(filtration.ndim)

    @staticmethod
    def reverse_order():
        return False

    @staticmethod
    def degree():
        return 0

    @staticmethod
    def is_dual():
        return False

def adj_vectors(dim):
    ret = []
    for k in range(dim):
        for direction in [1, -1]:
            vect = np.zeros((dim,), dtype=int)
            vect[k] = direction
            ret.append(vect)
    return ret

def dual_adj_vectors(dim):
    return [
        np.array(v) for v in itertools.product([1, 0, -1], repeat=dim)
        if not all(val == 0 for val in v)
    ]

def on_boundary(at, shape):
    return any(i == 0 or i == j - 1 for (i, j) in zip(at, shape))

def out_of_area(at, shape):
    return any(i < 0 or i >= j for (i, j) in zip(at, shape))


if __name__ == "__main__":
    main()
